from schorle.component import Component
from schorle.element import div, span
from schorle.rendering_context import rendering_context


def test_c():
    class TestComponent(Component):
        def render(self):
            span()
            with div():
                div()
            with div():
                div()

    with rendering_context() as rc:
        TestComponent()

    assert len(rc.root._children) == 1



from typing import Callable

from fastapi.responses import HTMLResponse
from lxml import etree
from pydantic import Field

from schorle.component import Component
from schorle.element import body, div, head, link, meta, script, span, title
from schorle.rendering_context import rendering_context
from schorle.tags import HTMLTag
from schorle.text import text
from schorle.theme import Theme


class DevLoader(Component):
    def render(self):
        with div(classes="absolute bottom-4 right-4 hidden", element_id="dev_loader"):
            span(classes="loading loading-md loading-bars text-primary")


class Document(Component):
    page: Component
    title: str = "Schorle"
    theme: Theme = Theme.DARK
    with_dev_meta: bool = False
    extra_assets: Callable[..., None] | None = None
    lang: str = "en"
    with_tailwind: bool = True
    with_daisyui: bool = True
    with_dev_tools: bool = False
    daisyui_version: str = "4.7.2"
    body_attrs: dict[str, str] | None = Field(default_factory=dict)
    tag: HTMLTag = HTMLTag.HTML

    def __init__(self, **data):
        super().__init__(**data)
        self.element_id = None
        self.attrs = dict(lang=self.lang, **{"data-theme": self.theme})

    def render(self):
        with head():
            meta(charset="utf-8")
            meta(name="viewport", content="width=device-width, initial-scale=1.0")
            if self.with_dev_tools:
                meta(name="schorle-dev", content="true")
            with title():
                text(self.title)
            if self.with_dev_meta:
                meta(name="schorle-dev", content="true")

            link(href="/favicon.svg", rel="icon", type="image/svg+xml")
            if self.with_tailwind:
                script(src="https://cdn.tailwindcss.com")
            if self.with_daisyui:
                link(
                    href=f"https://cdn.jsdelivr.net/npm/daisyui@{self.daisyui_version}/dist/full.min.css",
                    rel="stylesheet",
                    type="text/css",
                )

            script(src="/_schorle/js/index.js", crossorigin="anonymous", defer="", **{"type": "module"})
            if self.extra_assets:
                self.extra_assets()

        with body(**self.body_attrs):
            self.page()

            if self.with_dev_tools:
                DevLoader()

    def to_response(self) -> HTMLResponse:
        with rendering_context(root=self) as rc:
            self.render()
        return HTMLResponse(
            etree.tostring(rc.to_lxml(), pretty_print=True).decode("utf-8"),
            200,
        )



from lxml import etree

from schorle.element import div, span
from schorle.rendering_context import rendering_context


def test_one():
    with rendering_context() as rc:
        div()

    assert len(rc.root._children) == 1


def test_multi_exit():
    with rendering_context() as rc:
        with div():
            with div():
                div()
            with span():
                div()
        div()

    rendered = etree.tostring(rc.to_lxml(), pretty_print=False).decode("utf-8")
    assert rendered == "<root><div><div><div></div></div><span><div></div></span></div><div></div></root>"



from __future__ import annotations

from pydantic import BaseModel, Field, PrivateAttr

from schorle.tags import HTMLTag


class ElementPrototype(BaseModel):
    tag: HTMLTag | str
    element_id: str | None = None
    _children: list[ElementPrototype] = PrivateAttr(default_factory=list)
    _text: str | None = PrivateAttr(default=None)
    attrs: dict[str, str] = Field(default_factory=dict)
    classes: str | list[str] | None = None
    style: dict[str, str] | None = None

    def append(self, element: ElementPrototype):
        self._children.append(element)

    def set_text(self, text: str):
        self._text = text

    def get_children(self):
        return self._children

    def walk(self):
        for child in self._children:
            yield child
            yield from child.walk()

    def _cleanup(self):
        self._children = []
        self._text = None



from __future__ import annotations

from schorle.rendering_context import RENDERING_CONTEXT


def text(value: str):
    if not RENDERING_CONTEXT.get():
        raise RuntimeError("Text must be created inside a rendering context")
    else:
        rc = RENDERING_CONTEXT.get()
        rc.set_text(value)



import mimetypes
from importlib.resources import files
from pathlib import Path
from typing import Callable

from fastapi import FastAPI
from starlette.responses import FileResponse, HTMLResponse
from starlette.types import Receive, Scope, Send

from schorle.component import Component
from schorle.document import Document
from schorle.theme import Theme

ASSETS_PATH = Path(str(files("schorle"))) / Path("assets")


def favicon() -> FileResponse:
    favicon_path = ASSETS_PATH / "static" / "favicon.svg"
    return FileResponse(favicon_path, media_type="image/svg+xml")


def get_file(file_name: str) -> FileResponse:
    file_path = ASSETS_PATH / file_name
    mime_type, _ = mimetypes.guess_type(file_path)
    return FileResponse(file_path, media_type=mime_type)


SESSION_ID_HEADER = "X-Schorle-Session-Id"


class Schorle:
    def __init__(self, theme: Theme = Theme.DARK, lang: str = "en", extra_assets: Callable[..., None] | None = None):
        self.backend = FastAPI()
        self.backend.get("/_schorle/{file_name:path}", response_class=FileResponse)(get_file)
        self.backend.get("/favicon.svg", response_class=FileResponse)(favicon)
        self.theme = theme
        self.lang = lang
        self.extra_assets = extra_assets

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """
        This method is called by uvicorn when the server is started.
        """
        await self.backend(scope=scope, receive=receive, send=send)

    def get(self, path: str):
        def decorator(func: Callable[..., Component]):
            @self.backend.get(path, response_class=HTMLResponse)
            async def wrapper():
                doc = Document(page=func(), theme=self.theme, lang=self.lang, extra_assets=self.extra_assets)
                return doc.to_response()

        return decorator



from pathlib import Path

from starlette.responses import FileResponse, JSONResponse

from schorle.app import Schorle
from schorle.component import Component
from schorle.element import div, img, link, span
from schorle.icon import icon
from schorle.tags import HTMLTag
from schorle.text import text


def extra_assets():
    link(href="https://fonts.googleapis.com/css?family=Space+Mono", rel="stylesheet", type="text/css")


app = Schorle(extra_assets=extra_assets)


class LinkWithIcon(Component):
    tag: HTMLTag = HTMLTag.A
    href: str
    icon: str
    text: str
    classes: str = "btn btn-primary font-normal w-42"

    def initialize(self):
        self.attrs["href"] = self.href

    def render(self):
        icon(self.icon)
        with span():
            text(self.text)


LINKS = [
    LinkWithIcon(href="https://github.com/renardeinside/schorle", icon="github", text="GitHub"),
    LinkWithIcon(href="https://github.com/renardeinside/schorle/tree/main/examples", icon="code", text="Examples"),
    LinkWithIcon(
        href="https://medium.com/@polarpersonal/schorle-testing-the-waters-with-a-python-server-driven-ui-kit-053f85ee6574",
        icon="book-marked",
        text="Concepts",
    ),
]


class LandingPage(Component):
    def render(self):
        with div(classes="flex flex-col justify-center items-center h-screen", style={"font-family": "Space Mono"}):
            with div(classes="max-w-screen-md flex flex-col justify-center items-center space-y-4"):
                img(
                    src="/logo",
                    alt="Schorle logo",
                    media_type="image/svg+xml",
                    classes="w-3/4",
                )
            with span(classes="text-2xl m-4 p-4 w-5/6 text-center"):
                text("Pythonic Server-Driven UI Kit for building modern apps.")

            with div(classes="flex flex-col space-y-4 space-x-0 md:flex-row md:space-x-4 md:space-y-0"):
                for link_item in LINKS:
                    link_item()


@app.get("/")
def landing_page():
    return LandingPage()


@app.backend.get("/health", response_class=JSONResponse)
def health():
    return {"status": "ok"}


@app.backend.get("/logo", response_class=FileResponse)
def logo():
    return FileResponse(path=Path(__file__).parent.parent / Path("raw/with_text.svg"), media_type="image/svg+xml")



from __future__ import annotations

from schorle.prototypes import ElementPrototype
from schorle.rendering_context import RENDERING_CONTEXT
from schorle.tags import HTMLTag


class Element(ElementPrototype):

    def __init__(self, **data):
        super().__init__(**data)
        if not RENDERING_CONTEXT.get():
            raise RuntimeError("Element must be created inside a rendering context")
        else:
            RENDERING_CONTEXT.get().append(self)

    def __enter__(self):
        RENDERING_CONTEXT.get().become_parent(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        RENDERING_CONTEXT.get().reset_parent()

    def __repr__(self):
        return f"<{self.tag}>"

    def __str__(self):
        return self.__repr__()


def element_function_factory(tag: HTMLTag):
    def func(style: dict[str, str] | None = None, **kwargs):
        return Element(
            tag=tag,
            style=style,
            attrs=kwargs,
        )

    return func


div = element_function_factory(HTMLTag.DIV)
span = element_function_factory(HTMLTag.SPAN)
button = element_function_factory(HTMLTag.BUTTON)
input_ = element_function_factory(HTMLTag.INPUT)
a = element_function_factory(HTMLTag.A)
img = element_function_factory(HTMLTag.IMG)
script = element_function_factory(HTMLTag.SCRIPT)
link = element_function_factory(HTMLTag.LINK)
meta = element_function_factory(HTMLTag.META)
title = element_function_factory(HTMLTag.TITLE)
head = element_function_factory(HTMLTag.HEAD)
body = element_function_factory(HTMLTag.BODY)
html = element_function_factory(HTMLTag.HTML)
p = element_function_factory(HTMLTag.P)
h1 = element_function_factory(HTMLTag.H1)
h2 = element_function_factory(HTMLTag.H2)
h3 = element_function_factory(HTMLTag.H3)
h4 = element_function_factory(HTMLTag.H4)
h5 = element_function_factory(HTMLTag.H5)
h6 = element_function_factory(HTMLTag.H6)
ul = element_function_factory(HTMLTag.UL)
ol = element_function_factory(HTMLTag.OL)
li = element_function_factory(HTMLTag.LI)
table = element_function_factory(HTMLTag.TABLE)
tr = element_function_factory(HTMLTag.TR)
td = element_function_factory(HTMLTag.TD)
th = element_function_factory(HTMLTag.TH)
form = element_function_factory(HTMLTag.FORM)
footer = element_function_factory(HTMLTag.FOOTER)
icon = element_function_factory(HTMLTag.ICON)
main = element_function_factory(HTMLTag.MAIN)



from schorle.app import Schorle
from schorle.component import Component
from schorle.element import h1, span
from schorle.icon import icon
from schorle.text import text

app = Schorle()


class ExamplePage(Component):
    def render(self):
        with h1(classes="text-2xl w-96 flex justify-center items-center p-4"):
            icon(name="party-popper")
            with span():
                text("Welcome to Schorle!")


@app.get("/")
def home():
    return ExamplePage()



import sys
from enum import Enum

from schorle.types import LXMLElement


class RunningMode(str, Enum):
    DEV = "dev"
    PRODUCTION = "production"


def get_running_mode() -> RunningMode:
    # if uvicorn and --reload are in sys.argv, we are running in uvicorn dev mode
    _joined_argv = " ".join(sys.argv)
    if "uvicorn" in _joined_argv and "--reload" in _joined_argv:
        return RunningMode.DEV
    else:
        return RunningMode.PRODUCTION


def fix_self_closing_tags(element: LXMLElement) -> None:
    for _child in element.iter():
        if len(_child) == 0 and _child.text is None:
            _child.text = ""



from __future__ import annotations

from abc import abstractmethod
from uuid import uuid4

from lxml import etree

from schorle.prototypes import ElementPrototype
from schorle.rendering_context import RENDERING_CONTEXT, rendering_context
from schorle.tags import HTMLTag


class Component(ElementPrototype):
    tag: HTMLTag | str = HTMLTag.SCHORLE_COMPONENT

    def initialize(self):
        pass

    def __init__(self, **data):
        super().__init__(**data)

        if not self.element_id:
            self.element_id = f"sle-{str(uuid4())[0:8]}"
        self.initialize()

        self._append_to_context(strict=False)

    def _append_to_context(self, *, strict: bool = True):
        rc = RENDERING_CONTEXT.get()
        if strict and rc is None:
            raise RuntimeError("Component cannot be rendered outside of a RenderingContext")
        if rc is not None:
            RENDERING_CONTEXT.get().append(self)

    @abstractmethod
    def render(self):
        pass

    def _render(self):
        self._cleanup()
        with rendering_context(root=self) as rc:
            self.render()
            return rc.root

    def __call__(self):
        self._append_to_context(strict=True)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __str__(self):
        return self.__repr__()

    def to_string(self) -> str:
        with rendering_context(root=self) as rc:
            self.render()
        return etree.tostring(rc.to_lxml(), pretty_print=True).decode("utf-8")



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar

from lxml import etree

from schorle.prototypes import ElementPrototype
from schorle.tags import HTMLTag
from schorle.types import LXMLElement
from schorle.utils import fix_self_closing_tags


@contextmanager
def rendering_context(root: ElementPrototype | None = None):
    rc = RenderingContext(root=root)
    _token = RENDERING_CONTEXT.set(rc)
    yield rc
    RENDERING_CONTEXT.reset(_token)


class RenderingContext:
    def __init__(self, root: ElementPrototype | None = None):
        self.root = ElementPrototype(tag="root") if root is None else root
        self.current_parent = self.root
        self.previous_parent = None
        self.pre_previous_parent = None

    def append(self, element: ElementPrototype):
        self.current_parent.append(element)

    def become_parent(self, element: ElementPrototype):
        self.pre_previous_parent = self.previous_parent
        self.previous_parent = self.current_parent
        self.current_parent = element

    def reset_parent(self):
        self.current_parent = self.previous_parent
        self.previous_parent = self.pre_previous_parent

    def set_text(self, text_value: str):
        self.current_parent.set_text(text_value)

    def covert_proto_to_lxml(self, proto: ElementPrototype) -> LXMLElement:
        lxml_element = etree.Element(proto.tag.value if isinstance(proto.tag, HTMLTag) else proto.tag)

        if proto.element_id:
            lxml_element.set("id", proto.element_id)

        if proto._text:
            lxml_element.text = proto._text

        if proto.attrs:
            for key, value in proto.attrs.items():
                _key = "class" if key == "classes" else key
                lxml_element.set(_key, value)

        if proto.classes:
            if isinstance(proto.classes, str):
                proto.classes = [proto.classes]
            lxml_element.set("class", " ".join(proto.classes))

        if proto.style:
            lxml_element.set("style", ";".join([f"{key}: {value}" for key, value in proto.style.items()]))

        for child in proto.get_children():
            if hasattr(child, "_render"):
                rendered = child._render()
                lxml_child = self.covert_proto_to_lxml(rendered)
            else:
                lxml_child = self.covert_proto_to_lxml(child)
            lxml_element.append(lxml_child)

        fix_self_closing_tags(lxml_element)
        return lxml_element

    def to_lxml(self):
        return self.covert_proto_to_lxml(self.root)


RENDERING_CONTEXT: ContextVar[RenderingContext | None] = ContextVar("rendering_context", default=None)



import textwrap

from schorle.component import Component
from schorle.document import Document


def test_empty_doc():
    _title = "Test Document"

    class PlaceholderPage(Component):
        element_id: str = "schorle-page"

        def render(self):
            pass

    doc = Document(title=_title, page=PlaceholderPage())
    lnk = f"https://cdn.jsdelivr.net/npm/daisyui@{doc.daisyui_version}/dist/full.min.css"
    rendered = doc.to_string()
    expected = textwrap.dedent(
        f"""\
        <html lang="en" data-theme="dark">
          <head>
            <meta charset="utf-8"></meta>
            <meta name="viewport" content="width=device-width, initial-scale=1.0"></meta>
            <title>{_title}</title>
            <link href="/favicon.svg" rel="icon" type="image/svg+xml"></link>
            <script src="https://cdn.tailwindcss.com"></script>
            <link href="{lnk}" rel="stylesheet" type="text/css"></link>
            <script src="/_schorle/js/index.js" crossorigin="anonymous" defer="" type="module"></script>
          </head>
          <body>
            <schorle-component id="schorle-page"></schorle-component>
          </body>
        </html>
    """
    )
    assert rendered == expected
